package net.stoerr.sudoku;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Constraint solver for Sudoku.
 */
public class SudokuConstraintPropagator extends SudokuModel implements ConstraintStore<Integer> {
    private static final long serialVersionUID = -6316604934010862232L;
    private final Map<Integer, List<AbstractConstraint<Integer>>> pointToConstraintListMap = new HashMap<Integer, List<AbstractConstraint<Integer>>>();

    /*
     * (non-Javadoc)
     * @see net.stoerr.sudoku.SudokuModel#reset()
     */
    @Override
    public void reset() {
        super.reset();
        resetConstraints();
    }

    /**
     * Creates all necessary constraints.
     */
    protected void resetConstraints() {
        pointToConstraintListMap.clear();

        for (int r = 0; r < getRowCount(); ++r) {
            final AbstractConstraint<Integer> t = makeConstraint();

            for (int c = 0; c < getColumnCount(); ++c)
                t.addPoint(makeIndex(r, c));

            storeConstraint(t);
        }

        for (int c = 0; c < getColumnCount(); ++c) {
            final AbstractConstraint<Integer> t = makeConstraint();

            for (int r = 0; r < getRowCount(); ++r)
                t.addPoint(makeIndex(r, c));

            storeConstraint(t);
        }

        for (int rd = 0; rd < getRowCount(); rd += 3)
            for (int cd = 0; cd < getColumnCount(); cd += 3) {
                final AbstractConstraint<Integer> t = makeConstraint();

                for (int r = rd; r < rd + 3; ++r)
                    for (int c = cd; c < cd + 3; ++c)
                        t.addPoint(makeIndex(r, c));

                storeConstraint(t);
            }
    }

    /**
     * Stores a constraint in pointToConstraintListMap for each occurring point.
     * @param c
     */
    protected void storeConstraint(final AbstractConstraint<Integer> c) {
        for (final Integer p : c.getPoints()) {
            List<AbstractConstraint<Integer>> l = pointToConstraintListMap.get(p);

            if (null == l) {
                l = new ArrayList<AbstractConstraint<Integer>>();
                pointToConstraintListMap.put(p, l);
            }

            l.add(c);
        }
    }

    /**
     * Notifies the constraints of a change.
     * @param idx
     * @param val
     */
    private void propagateChange(final Integer idx, final String val) {
        final List<AbstractConstraint<Integer>> l = pointToConstraintListMap.get(idx);

        for (final AbstractConstraint<Integer> c : l)
            c.propagate(idx, val);
    }

    /**
     * Factory method for the used constraints.
     * @return
     */
    private AbstractConstraint<Integer> makeConstraint() {
        return new PermutationConstraint<Integer>(this);
    }

    /**
     * Sets the constraint and propagates the changes.
     * @see net.stoerr.sudoku.SudokuModel#setConstraint(java.lang.Integer, java.lang.String)
     */
    @Override
    public void setConstraint(final Integer idx, final String val) {
        super.setConstraint(idx, val);

        if (null == val || 0 == val.length()) haveConflict = true;
        else propagateChange(idx, val);
    }

    /**
     * Solves remainder by backtracking. We look for one of the fields with the least possible digits and try every
     * digit until we have a solution. This is an awful but sufficient implementation: undo resets and redoes everything
     * instead of saving the current state.
     * @return true if solved.
     */
    @Override
    public boolean solve() {
        Integer leastpos = null;
        String possibilities = "a string that is longer than 123456789";

        for (final Map.Entry<Integer, String> entry : values.entrySet()) {
            final Integer idx = entry.getKey();
            final String val = entry.getValue();

            if (1 < val.length() && possibilities.length() > val.length()) {
                leastpos = idx;
                possibilities = val;
            }
        }

        if (null == leastpos) return true; // everything solved or everything completely unclear.

        for (int i = 0; i < possibilities.length(); ++i) {
            final char trychar = possibilities.charAt(i);
            makeMove(leastpos, String.valueOf(trychar));

            if (!haveConflict) if (solve()) return true;

            undo();
        }

        return false;
    }
}
