package net.stoerr.sudoku;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.table.AbstractTableModel;

/**
 * A tablemodel displaying the sudoku. Integrates stub methods for solving etc., which are overridden by
 * implementations.
 */
public class SudokuModel extends AbstractTableModel {
    private static final long serialVersionUID = -4827881144666547990L;
    private static final int ROWS = 9;
    private static final int COLS = 9;
    public static final String FULLVALUES = "123456789";
    protected Map<Integer, String> values = new HashMap<Integer, String>();
    protected List<Integer> movePositions = new ArrayList<Integer>();
    protected List<String> moveConstraints = new ArrayList<String>();
    protected boolean haveConflict;

    /* @see javax.swing.table.TableModel#getColumnCount() */
    public final int getColumnCount() {
        return COLS;
    }

    /* @see javax.swing.table.TableModel#getRowCount() */
    public final int getRowCount() {
        return ROWS;
    }

    /* @see javax.swing.table.TableModel#getValueAt(int, int) */
    public final Object getValueAt(final int rowIndex, final int columnIndex) {
        return values.get(makeIndex(rowIndex, columnIndex));
    }

    /**
     * Remembers a hint.
     * @see javax.swing.table.AbstractTableModel#setValueAt(java.lang.Object, int, int)
     */
    @Override
    public final void setValueAt(final Object aValue, final int rowIndex, final int columnIndex) {
        final char c = ((Character) aValue).charValue();

        if (0 <= FULLVALUES.indexOf(c)) {
            final Integer idx = makeIndex(rowIndex, columnIndex);
            final String cS = String.valueOf(c);
            final String current = values.get(idx);
            if (null != cS && current.contains(cS) && !cS.equals(current)) makeMove(idx, cS);
        }
    }

    /**
     * Adds the given constraint to the model - from the human or the backtracker.
     * @param idx field to be constrained
     * @param val constraint
     */
    protected void makeMove(final Integer idx, final String val) {
        movePositions.add(idx);
        moveConstraints.add(val);
        setConstraint(idx, val);
    }

    /**
     * Adds a constraint to the model. Override this to implement incremental solver.
     * @param idx field to be constrained
     * @param val constraint
     */
    public void setConstraint(final Integer idx, final String val) {
        values.put(idx, val);
    }

    /**
     * Returns the constraint for the point.
     */
    public final String getConstraint(final Integer idx) {
        return values.get(idx);
    }

    /**
     * Constructs the integer used as representation for the point.
     * @param row
     * @param col
     * @return 10 * row + col;
     */
    protected final Integer makeIndex(final int row, final int col) {
        return 10 * row + col;
    }

    /**
     * Full scale reset.
     */
    public void reset() {
        values.clear();
        movePositions.clear();
        moveConstraints.clear();
        haveConflict = false;

        for (int i = 0; i < ROWS; ++i)
            for (int j = 0; j < COLS; ++j) {
                final Integer idx = makeIndex(i, j);
                values.put(idx, FULLVALUES);
            }
    }

    /**
     * Undo last move. This is done the hard way - resetting everything and redoing except last move.
     */
    public final void undo() {
        final List<Integer> redoPositions = new ArrayList<Integer>(movePositions);
        final List<String> redoConstraints = new ArrayList<String>(moveConstraints);
        reset();

        for (int i = 0; i < redoPositions.size() - 1; ++i)
            makeMove(redoPositions.get(i), redoConstraints.get(i));
    }

    /**
     * Stub method for solving e.g. by backtracking, to be overridden.
     * @return if we found a solution
     */
    public boolean solve() {
        // does nothing
        return false;
    }
}
